/* 
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

Title : OpenCL implementation of FDK Algorithm

Authors : Atapattu, Chatura
		  Houck, Dustin
		  O'Brien, Ronald
		  Partel, Michael

Description : This is an OpenCL implementation of the FDK
			  algorithm used to convert 2D cone-beam images
			  into a 3D volumetric image

Created : 11/01/2009

Modified : ongoing

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
*/

// Define constants
#define _CRT_SECURE_NO_DEPRECATE
#define READ_PFM
#define WRITE_BLOCK (1024*1024)
#define DEFAULT_WARP_SIZE 6
#define MAX_GPU_COUNT 8

// Include project header
#include "fdk_opencl.h"

// Globals
cl_int error;					// Use for error checking
cl_context context;				// Context from device
cl_command_queue command_queue[MAX_GPU_COUNT];	// Command Queue from Context
cl_program program;				// Program from .cl file
cl_uint compute_units[MAX_GPU_COUNT];			// CL_DEVICE_MAX_COMPUTE_UNITS
size_t workitem_dims[MAX_GPU_COUNT];			// CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS
size_t workitem_size[MAX_GPU_COUNT][3];		// CL_DEVICE_MAX_WORK_ITEM_SIZES
size_t workgroup_size[MAX_GPU_COUNT];			// CL_DEVICE_MAX_WORK_GROUP_SIZE
cl_uint warp_size[MAX_GPU_COUNT];				// CL_NV_DEVICE_WARP_SIZE
cl_uint device_count;			// number of devices available
cl_device_id* devices;			// pointer to devices
cl_device_id device;			// object for individual device in for loop


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
cl_ulong executionTime(cl_event &event)
{
    cl_ulong start, end;
    
    clGetEventProfilingInfo(event, CL_PROFILING_COMMAND_END, sizeof(cl_ulong), &end, NULL);
    clGetEventProfilingInfo(event, CL_PROFILING_COMMAND_START, sizeof(cl_ulong), &start, NULL);
   
    return (end - start);
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void free_cb_image (CB_Image* cbi)
{
	free (cbi->img);
	free (cbi);
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void fwrite_block (void* buf, size_t size, size_t count, FILE* fp)
{
	size_t left_to_write = count * size;
	size_t cur = 0;
	char* bufc = (char*) buf;

	while (left_to_write > 0) {
	size_t this_write, rc;

	this_write = left_to_write;
	if (this_write > WRITE_BLOCK) this_write = WRITE_BLOCK;
	rc = fwrite (&bufc[cur], 1, this_write, fp);
	if (rc != this_write) {
		shrLog("Error writing to file. rc=%zd, this_write=%zd\n", rc, this_write);
		return;
	}
	cur += rc;
	left_to_write -= rc;
	}
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void parse_args (MGHCBCT_Options* options, int argc, char* argv[])
{
	int i, rc;

	if (argc < 2)
	{ print_usage(); exit(1); }

	set_default_options (options);
	for (i = 1; i < argc; i++) {
		if (argv[i][0] != '-') break;
		if (!strcmp (argv[i], "-r")) {
			if (i == (argc-1) || argv[i+1][0] == '-') {
				shrLog("option %s requires an argument\n", argv[i]);
				exit(1);
			}
			i++;
			rc = sscanf (argv[i], "%d %d %d", 
			&options->resolution[0], 
			&options->resolution[1],
			&options->resolution[2]);
			if (rc == 1) {
				options->resolution[1] = options->resolution[0];
				options->resolution[2] = options->resolution[0];
			} else if (rc != 3) {
				print_usage ();
			}
		} else if (!strcmp (argv[i], "-I")) {
			if (i == (argc-1) || argv[i+1][0] == '-') {
				shrLog("option %s requires an argument\n", argv[i]);
				exit(1);
			}
			i++;
			options->input_dir = strdup(argv[i]);
		} else if (!strcmp (argv[i], "-O")) {
			if (i == (argc-1) || argv[i+1][0] == '-') {
				shrLog("option %s requires an argument\n", argv[i]);
				exit(1);
			}
			i++;
			options->output_file = strdup(argv[i]);
		} else if (!strcmp (argv[i], "-a")) {
			if (i == (argc-1) || argv[i+1][0] == '-') {
				shrLog("option %s requires an argument\n", argv[i]);
				exit(1);
			}
			i++;
			rc = sscanf (argv[i], "%d %d %d" , 
			&options->first_img,
			&options->skip_img,
			&options->last_img);
			if (rc == 1) {
				options->last_img = options->first_img;
				options->skip_img = 1;
			} else if (rc == 2) {
				options->last_img = options->skip_img;
				options->skip_img = 1;
			} else if (rc != 3) {
				print_usage ();
			}
		} else if (!strcmp (argv[i], "-s")) {
			if (i == (argc-1) || argv[i+1][0] == '-') {
				shrLog("option %s requires an argument\n", argv[i]);
				exit(1);
			}
			i++;
			rc = sscanf (argv[i], "%g" , &options->scale);
			if (rc != 1) {
				print_usage ();
			}
		} else if (!strcmp (argv[i], "-z")) {
			if (i == (argc-1) || argv[i+1][0] == '-') {
				shrLog("option %s requires an argument\n", argv[i]);
				exit(1);
			}
			i++;
			rc = sscanf (argv[i], "%g %g %g", 
			&options->vol_size[0],
			&options->vol_size[1],
			&options->vol_size[2]);
			if (rc == 1) {
				options->vol_size[1] = options->vol_size[0];
				options->vol_size[2] = options->vol_size[0];
			} else if (rc != 3) {
				print_usage ();
			}
		} else {
			print_usage ();
		}
	}
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void print_usage (void)
{
	shrLog("Usage: mghcbct [options]\n"
		"Options:\n"
		" -a \"num ((num) num)\"		   Use this range of images\n"
		" -r \"r1 r2 r3\"				  Set output resolution (in voxels)\n"
		" -s scale					   Scale the intensity of the output file\n"
		" -z \"s1 s2 s3\"				  Physical size of the reconstruction (in mm)\n"
		" -I indir					   The input directory\n"
		" -O outfile					 The output file\n"
		);
	exit(1);
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void set_default_options (MGHCBCT_Options* options)
{
	options->first_img = 0;
	options->last_img = 119;
	options->skip_img = 1;
	options->resolution[0] = 120;
	options->resolution[1] = 120;
	options->resolution[2] = 120;
	options->vol_size[0] = 500.0f;
	options->vol_size[1] = 500.0f;
	options->vol_size[2] = 500.0f;
	options->scale = 1.0f;
	options->input_dir = ".";
	options->output_file = "output.mha";
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void write_mha (const char* filename, Volume* vol)
{
	FILE* fp;
	const char* mha_header = 
		"ObjectType = Image\n"
		"NDims = 3\n"
		"BinaryData = True\n"
		"BinaryDataByteOrderMSB = False\n"
		"TransformMatrix = 1 0 0 0 1 0 0 0 1\n"
		"Offset = %g %g %g\n"
		"CenterOfRotation = 0 0 0\n"
		"ElementSpacing = %g %g %g\n"
		"DimSize = %d %d %d\n"
		"AnatomicalOrientation = RAI\n"
		"%s"
		"ElementType = %s\n"
		"ElementDataFile = LOCAL\n";
	const char* element_type;

	if (vol->pix_type == PT_VF_FLOAT_PLANAR) {
		shrLog("Error, PT_VF_FLOAT_PLANAR not implemented\n");
		exit (-1);
	}

	fp = fopen (filename,"wb");
	if (!fp) {
		shrLog("Can't open file %s for write\n", filename);
		return;
	}
	switch (vol->pix_type) {
		case PT_UCHAR:
			element_type = "MET_UCHAR";
			break;
		case PT_SHORT:
			element_type = "MET_SHORT";
			break;
		case PT_UINT32:
			element_type = "MET_UINT";
			break;
		case PT_FLOAT:
			element_type = "MET_FLOAT";
			break;
		case PT_VF_FLOAT_INTERLEAVED:
			element_type = "MET_FLOAT";
			break;
		case PT_VF_FLOAT_PLANAR:
		default:
			shrLog("Unhandled type in write_mha().\n");
			exit (-1);
	}
	fprintf (fp, mha_header, 
		vol->offset[0], vol->offset[1], vol->offset[2], 
		vol->pix_spacing[0], vol->pix_spacing[1], vol->pix_spacing[2], 
		vol->dim[0], vol->dim[1], vol->dim[2],
		(vol->pix_type == PT_VF_FLOAT_INTERLEAVED) 
		? "ElementNumberOfChannels = 3\n" : "",
		element_type);
	fflush (fp);

	fwrite_block (vol->img, vol->pix_size, vol->npix, fp);

	fclose (fp);
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
CB_Image* get_image (MGHCBCT_Options* options, int image_num)
{
#if defined (READ_PFM)
	const char* img_file_pat = "out_%04d.pfm";
#else
	const char* img_file_pat = "out_%04d.pgm";
#endif
	const char* mat_file_pat = "out_%04d.txt";

	char img_file[1024], mat_file[1024], fmt[1024];
	sprintf(fmt, "%s/%s", options->input_dir, img_file_pat);
	sprintf(img_file, fmt, image_num);
	sprintf(fmt, "%s/%s", options->input_dir, mat_file_pat);
	sprintf(mat_file, fmt, image_num);
	return load_cb_image(img_file, mat_file);
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
CB_Image* load_cb_image (char *img_filename, char *mat_filename)
{
	int i;
	size_t rc;
	float f;
	FILE *fp;
	char buf[1024];
	CB_Image* cbi;

	cbi = (CB_Image*)malloc(sizeof(CB_Image));

	fp = fopen (img_filename,"rb");
	if (!fp) {
		shrLog("Can't open file %s for read\n", img_filename);
		exit(-1);
	}

	#if defined (READ_PFM)
		/* Verify that it is pfm */
		fgets(buf, 1024, fp);
		if (strncmp(buf, "Pf", 2)) {
			shrLog("Couldn't parse file %s as an image [1]\n", img_filename);
			printf (buf);
			exit(-1);
		}

		/* Get image resolution */
		fgets(buf, 1024, fp);
		if (2 != sscanf (buf, "%d %d", &cbi->dim[0], &cbi->dim[1])) {
			shrLog("Couldn't parse file %s as an image [2]\n", img_filename);
			exit(-1);
		}
		/* Skip third line */
		fgets(buf, 1024, fp);

		/* Malloc memory */
		cbi->img = (float*) malloc (sizeof(float) * cbi->dim[0] * cbi->dim[1]);
		if (!cbi->img) {
			shrLog("Couldn't malloc memory for input image\n");
			exit(-1);
		}

		/* Load pixels */
		rc = fread(cbi->img, sizeof(float), cbi->dim[0] * cbi->dim[1], fp);
		if (rc != (size_t)(cbi->dim[0] * cbi->dim[1])) {
			shrLog("Couldn't load raster data for %s\n", img_filename);
			exit (-1);
		}
	#else
		/* Verify that it is pgm */
		fgets(buf, 1024, fp);
		if (strncmp(buf, "P2", 2)) {
			shrLog("Couldn't parse file %s as an image [1]\n", img_filename);
			printf (buf);
			exit (-1);
		}

		/* Skip comment line */
		fgets(buf, 1024, fp);

		/* Get image resolution */
		fgets (buf, 1024, fp);
		if (2 != sscanf (buf, "%d %d", &cbi->dim[0], &cbi->dim[1])) {
			shrLog("Couldn't parse file %s as an image [2]\n", img_filename);
			exit (-1);
		}

		/* Skip max gray */
		fgets(buf, 1024, fp);

		/* Malloc memory */
		cbi->img = (float*) malloc (sizeof(float) * cbi->dim[0] * cbi->dim[1]);
		if (!cbi->img) {
			shrLog("Couldn't malloc memory for input image\n");
			exit (-1);
		}

		/* Load pixels */
		for (i = 0; i < cbi->dim[0] * cbi->dim[1]; i++) {
			if (1 != fscanf (fp, "%g", &cbi->img[i])) {
				shrLog("Couldn't parse file %s as an image [3,%d]\n", img_filename, i);
				exit (-1);
			}
		}
	#endif
	fclose(fp);

	/* Load projection matrix */
	fp = fopen (mat_filename,"r");
	if (!fp) {
		shrLog("Can't open file %s for read\n", mat_filename);
		exit(-1);
	}

	/* Load image center */
	for (i = 0; i < 2; i++) {
		if (1 != fscanf (fp, "%g", &f)) {
			shrLog("Couldn't parse file %s as a matrix [1,%d]\n", mat_filename, i);
			exit(-1);
		}
		cbi->ic[i] = (double) f;
	}

	/* Load projection matrix */
	for (i = 0; i < 12; i++) {
		if (1 != fscanf (fp, "%g", &f)) {
			shrLog("Couldn't parse file %s as a matrix [2,%d]\n", mat_filename, i);
			exit (-1);
		}
		cbi->matrix[i] = (double) f;
	}

	/* Load sad */
	if (1 != fscanf (fp, "%g", &f)) {
		shrLog("Couldn't load sad from %s\n", mat_filename);
		exit (-1);
	}
	cbi->sad = (double) f;

	/* Load sid */
	if (1 != fscanf (fp, "%g", &f)) {
		shrLog("Couldn't load sad from %s\n", mat_filename);
		exit (-1);
	}
	cbi->sid = (double) f;

	/* Load nrm vector */
	for (i = 0; i < 3; i++) {
		if (1 != fscanf (fp, "%g", &f)) {
			shrLog("Couldn't parse file %s as a matrix [1,%d]\n", mat_filename, i);
			exit (-1);
		}
		cbi->nrm[i] = (double) f;
	}
	fclose (fp);

	return cbi;
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
Volume* my_create_volume (MGHCBCT_Options* options)
{
	float offset[3];
	float spacing[3];
	float* vol_size = options->vol_size;
	int* resolution = options->resolution;

	spacing[0] = vol_size[0] / resolution[0];
	spacing[1] = vol_size[1] / resolution[1];
	spacing[2] = vol_size[2] / resolution[2];

	offset[0] = -vol_size[0] / 2.0f + spacing[0] / 2.0f;
	offset[1] = -vol_size[1] / 2.0f + spacing[1] / 2.0f;
	offset[2] = -vol_size[2] / 2.0f + spacing[2] / 2.0f;

	return volume_create (resolution, offset, spacing, PT_FLOAT, 0, 0);
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
Volume* volume_create (int *dim, float *offset, float *pix_spacing, enum Pixel_Type pix_type, float *direction_cosines,	int min_size)
{
	int i;
	Volume* vol = (Volume*) malloc (sizeof(Volume));
	if (!vol) {
		shrLog("Memory allocation failed.\n");
		exit(1);
	}

	memset (vol, 0, sizeof(Volume));
	for (i = 0; i < 3; i++) {
		vol->dim[i] = dim[i];
		vol->offset[i] = offset[i];
		vol->pix_spacing[i] = pix_spacing[i];
	}
	if (direction_cosines) {
		memcpy (vol->direction_cosines, direction_cosines, sizeof(vol->direction_cosines));
	} else {
		vol->direction_cosines[0] = 1.0f;
		vol->direction_cosines[4] = 1.0f;
		vol->direction_cosines[8] = 1.0f;
	}
	vol->npix = vol->dim[0] * vol->dim[1] * vol->dim[2];
	vol->pix_type = pix_type;

	switch (pix_type) {
		case PT_UCHAR:
			vol->pix_size = sizeof(unsigned char);
			break;
		case PT_SHORT:
			vol->pix_size = sizeof(short);
			break;
		case PT_FLOAT:
			vol->pix_size = sizeof(float);
			break;
		case PT_VF_FLOAT_INTERLEAVED:
			vol->pix_size = 3 * sizeof(float);
			break;
		case PT_VF_FLOAT_PLANAR:
			vol->pix_size = sizeof(float);
			break;
		default:
			shrLog("Unhandled type in volume_create().\n");
			exit (-1);
	}

	if (pix_type == PT_VF_FLOAT_PLANAR) {
		int i;
		int alloc_size = min_size;
		float** der = (float**) malloc (3*sizeof(float*));
		if (!der) {
			shrLog("Memory allocation failed.\n");
			exit(1);
		}
		if (alloc_size < vol->npix) alloc_size = vol->npix;
		shrLog("alloc_size = %d, min_size = %d, npix = %d\n", alloc_size, min_size, vol->npix);
		for (i=0; i < 3; i++) {
			der[i] = (float*) malloc (alloc_size*sizeof(float));
			if (!der[i]) {
				shrLog("Memory allocation failed.\n");
				exit(1);
			}
			memset(der[i], 0, alloc_size*sizeof(float));
		}
		vol->img = (void*) der;
	} else {
		vol->img = (void*) malloc (vol->pix_size * vol->npix);
		if (!vol->img) {
			shrLog("Memory allocation failed (alloc size = %d).\n",
			vol->pix_size * vol->npix);
			exit(1);
		}
		memset (vol->img, 0, vol->pix_size * vol->npix);
	}

	return vol;
}


/*
 * Summary: Gets the device warp if device is an NVIDIA graphics card		
 * Parameters: device id
 * Return: cl_uint of the device warp
 */
cl_uint get_device_warp(cl_device_id device) {
	char device_string[1024];
	bool nv_device_attibute_query = false;
	
	// CL_DEVICE_EXTENSIONS: get device extensions, and if any then parse & log the string onto separate lines
	clGetDeviceInfo(device, CL_DEVICE_EXTENSIONS, sizeof(device_string), &device_string, NULL);
	if (device_string != 0) 
	{
		std::string stdDevString;
		stdDevString = std::string(device_string);
		size_t szOldPos = 0;
		size_t szSpacePos = stdDevString.find(' ', szOldPos); // extensions string is space delimited
		while (szSpacePos != stdDevString.npos && (szSpacePos - szOldPos) > 0)
		{
			if( strcmp("cl_nv_device_attribute_query", stdDevString.substr(szOldPos, szSpacePos - szOldPos).c_str()) == 0 )
				nv_device_attibute_query = true;
			
			szOldPos = szSpacePos + 1;
			szSpacePos = stdDevString.find(' ', szOldPos); 
		}
	}

	cl_uint temp_warp_size;

	if(nv_device_attibute_query) 
	{
		clGetDeviceInfo(device, CL_DEVICE_WARP_SIZE_NV, sizeof(cl_uint), &temp_warp_size, NULL);
	} else {
		temp_warp_size = DEFAULT_WARP_SIZE;
	}

	return temp_warp_size;
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */

int main(int argc, char* argv[])
{
	// Initialize variables
	float total_runtime;

	// Set logfile name and start logs
	shrSetLogFileName ("fdk_opencl.txt");

	// Output command call
	shrLog("\n\n");
	for (int i = 0; i < argc; i++)
		shrLog("%s ", argv[i]);
	shrLog("\n\nStarting FDK_OPENCL...\n\n"); 

	/**************************************************************** 
	 * STEP 1: Setup OpenCL										  * 
	 ****************************************************************/

	// Used for error checking
	cl_int error = 0;

	// Get the OpenCL platform
	cl_platform_id platform;
	error = oclGetPlatformID(&platform);
	oclCheckError(error, CL_SUCCESS);

	// get devices of type GPU
	error = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 0, NULL, &device_count);
	oclCheckError(error, CL_SUCCESS);

	// makes sure using no more than the maximum number of GPUs
	if(device_count > MAX_GPU_COUNT)
		device_count = MAX_GPU_COUNT;

	devices = (cl_device_id *)malloc(device_count * sizeof(cl_device_id) );
	error = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, device_count, devices, NULL);
	oclCheckError(error, CL_SUCCESS);

	// create context
	cl_context_properties properties[] = {CL_CONTEXT_PLATFORM, (cl_context_properties)platform, 0};
	context = clCreateContext(properties, device_count, devices, NULL, NULL, &error);
	oclCheckError(error, CL_SUCCESS);

	shrLog("Using %d device(s):\n", device_count);
	char device_name[256];

	for (cl_uint i = 0; i < device_count; i++) {
		// device info
		device = oclGetDev(context, i);
		clGetDeviceInfo(device, CL_DEVICE_NAME, sizeof(device_name), device_name, NULL);
		clGetDeviceInfo(device, CL_DEVICE_MAX_COMPUTE_UNITS, sizeof(compute_units), &compute_units[i], NULL);
		clGetDeviceInfo(device, CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS, sizeof(workitem_dims), &workitem_dims[i], NULL);
		clGetDeviceInfo(device, CL_DEVICE_MAX_WORK_ITEM_SIZES, sizeof(workitem_size), &workitem_size[i], NULL);
		clGetDeviceInfo(device, CL_DEVICE_MAX_WORK_GROUP_SIZE, sizeof(workgroup_size), &workgroup_size[i], NULL);
		warp_size[i] = get_device_warp(device);
		oclCheckError(error, CL_SUCCESS);
		shrLog("\tDevice %d: %s\n", i, device_name);

		// command queue
		command_queue[i] = clCreateCommandQueue(context, device, CL_QUEUE_PROFILING_ENABLE, &error);
		oclCheckError(error, CL_SUCCESS);
	}

	// Program Setup
	size_t program_length;
	char* source_path = shrFindFilePath("fdk_opencl.cl", argv[0]);
	oclCheckError(source_path != NULL, shrTRUE);
	char *source = oclLoadProgSource(source_path, "", &program_length);
	oclCheckError(source != NULL, shrTRUE);

	// Create the program
	program = clCreateProgramWithSource(context, 1, (const char **)&source, &program_length, &error);
	oclCheckError(error, CL_SUCCESS);

	// Build the program
	error = clBuildProgram(program, 0, NULL, NULL, NULL, NULL);
	if (error != CL_SUCCESS) {
		// write out standard error, Build Log and PTX, then return error
		shrLogEx(LOGBOTH | ERRORMSG, error, STDERROR);
		oclLogBuildInfo(program, oclGetFirstDev(context));
		oclLogPtx(program, oclGetFirstDev(context), "fdk_opencl.ptx");
		return(EXIT_FAILURE); 
	}

	/****************************************************************/

	// Begin FDK Implementation
	MGHCBCT_Options options;
	Volume *vol;

	////////////////////////////////////
	#if defined (READ_PFM)
		shrLog("\n[PFM Input]\n");
	#else
		shrLog("\n[PGM Input]\n"); 
	#endif
	////////////////////////////////////

	/**************************************************************** 
	 * STEP 1: Parse commandline arguments						   * 
	 ****************************************************************/
	parse_args(&options, argc, argv);
	
	/*****************************************************
	 * STEP 2: Create the 3D array of voxels			  *
	 *****************************************************/
	vol = my_create_volume(&options);

	 /********************************************************************
	 * STEP 3: Reconstruct/populate the volume, convert to HU values	  *
	 *********************************************************************/
	total_runtime = reconstruct_conebeam_and_convert_to_hu(vol, &options);

	/*************************************
	 * STEP 4: Write MHA output file	  *
	 *************************************/
	shrLog("Writing output volume...");
	write_mha (options.output_file, vol);

	/*************************************
	 * STEP 5: Cleanup OpenCL and finish  *
	 *************************************/
	// Cleanup opencl
	clReleaseProgram(program);
	for (cl_uint i=0; i < device_count; i++)
		clReleaseCommandQueue(command_queue[i]);

	clReleaseContext(context);

	shrLog("\n\nDone FDK_OPENCL...\n\n");

	shrLog("Total OpenCL run time: %f s\n", total_runtime);

	return 0;
}


/*
 * Summary:		This OpenCL stub function performs most of the IO and flow
				control.  The actual back projection is performed by the
				OpenCL kernel "fdk_kernel()"
 * Parameters:	Pointer to volume and runtime options
 * Return:		Float of runtime
 */
float reconstruct_conebeam_and_convert_to_hu (Volume *vol, MGHCBCT_Options *options)
{
	int4 voxels_per_device[MAX_GPU_COUNT];
	int4 vpd;
	int4 extra;
	int4 offset[MAX_GPU_COUNT];

	// Calculate number of voxels per device
	vpd.x = vol->dim[0];
	vpd.y = vol->dim[1];
	vpd.z = vol->dim[2]/device_count;

	// Calculate any extra voxels allotted to last device
	extra.x = vol->dim[0] % vpd.x;
	extra.y = vol->dim[1] % vpd.y;
	extra.z = vol->dim[2] % vpd.z;

	// Allocate voxels to each device 
	for (cl_uint i = 0; i < device_count; i++) {
		voxels_per_device[i].x = vpd.x;
		voxels_per_device[i].y = vpd.y;
		voxels_per_device[i].z = vpd.z;
		offset[i].x = 0;
		offset[i].y = 0;
		offset[i].z = i * vpd.z;
	}

	// Allocate any extra voxels to the last device
	voxels_per_device[device_count - 1].x += extra.x;
	voxels_per_device[device_count - 1].y += extra.y;
	voxels_per_device[device_count - 1].z += extra.z;

	for (cl_uint i = 0; i < device_count; i++) {
		voxels_per_device[i].w = voxels_per_device[i].x * voxels_per_device[i].y * voxels_per_device[i].z;
		offset[i].w = offset[i].x + offset[i].y * vol->dim[0] + offset[i].z * vol->dim[0] * vol->dim[1];
	}

	//////////////////// TODO: Optimize threads in groups/cores ////////////////////
	size_t fdk_local_work_size[MAX_GPU_COUNT][3];
	size_t fdk_global_work_size[MAX_GPU_COUNT][3];

	for (cl_uint i=0; i < device_count; i++) {
		// Thead Block Dimensions (local work size)
		int fdk_lBlock_x = 512;
		int fdk_lBlock_y = 1;
		int fdk_lBlock_z = 1;

		// Work group dimensions (global work size)
		int fdk_gBlock_x = voxels_per_device[i].x + (fdk_lBlock_x - (voxels_per_device[i].x & (fdk_lBlock_x - 1)));
		int fdk_gBlock_y = voxels_per_device[i].y + (fdk_lBlock_y - (voxels_per_device[i].y & (fdk_lBlock_y - 1)));
		int fdk_gBlock_z = voxels_per_device[i].z + (fdk_lBlock_z - (voxels_per_device[i].z & (fdk_lBlock_z - 1)));

		// Each element in the volume (each voxel) gets 1 thread
		fdk_local_work_size[i][0] = fdk_lBlock_x;
		fdk_local_work_size[i][1] = fdk_lBlock_y;
		fdk_local_work_size[i][2] = fdk_lBlock_z;
		fdk_global_work_size[i][0] = fdk_gBlock_x;
		fdk_global_work_size[i][1] = fdk_gBlock_y;
		fdk_global_work_size[i][2] = fdk_gBlock_z;
	}
	////////////////////////////////////////////////////////////////////////////////

	// Structure for passing arugments to kernel: (See fdk_cuda.h)
	kernel_args_fdk *kargs;
	kargs = (kernel_args_fdk *) malloc(sizeof(kernel_args_fdk));

	// Calculate the scale
	int image_num = 1 + (options->last_img - options->first_img) / options->skip_img;
	float scale = (float)(sqrt(3.0)/(double)image_num);
	scale = scale * options->scale;

	// Load static kernel arguments
	kargs->scale = scale;
	kargs->vol_offset.x = vol->offset[0];
	kargs->vol_offset.y = vol->offset[1];
	kargs->vol_offset.z = vol->offset[2];
	kargs->vol_dim.x = vol->dim[0];
	kargs->vol_dim.y = vol->dim[1];
	kargs->vol_dim.z = vol->dim[2];
	kargs->vol_pix_spacing.x = vol->pix_spacing[0];
	kargs->vol_pix_spacing.y = vol->pix_spacing[1];
	kargs->vol_pix_spacing.z = vol->pix_spacing[2];

	// State the kernel execution parameters
	//shrLog("kernel parameters:\n dimGrid: %u, %u (Logical: %u, %u, %u)\n dimBlock: %u, %u, %u\n", dimGrid.x, dimGrid.y, dimGrid.x, blocksInY, blocksInZ, dimBlock.x, dimBlock.y, dimBlock.z);
	shrLog("%u voxels in volume\n", vol->npix);
	shrLog("%u projections to process\n", 1+(options->last_img - options->first_img) / options->skip_img);
	shrLog("%u Total Operations\n", vol->npix * (1+(options->last_img - options->first_img) / options->skip_img));
	shrLog("========================================\n\n");

	// Initiate static size of memory buffers
	int matrix_size = 12 * sizeof(float);

	// Retrieve 2D image to get dimensions 
	CB_Image *cbi = get_image(options, options->first_img);

	if (cbi->dim[0] > CL_DEVICE_IMAGE2D_MAX_WIDTH || cbi->dim[1] > CL_DEVICE_IMAGE2D_MAX_HEIGHT) {
		shrLog("Image dimensions too large for %s\n", CL_DEVICE_NAME);
		shrLog("Maximum height and width: %d x %d\n", CL_DEVICE_IMAGE2D_MAX_HEIGHT, CL_DEVICE_IMAGE2D_MAX_WIDTH);
		shrLog("Image height and width: %d x %d\n", cbi->dim[1], cbi->dim[0]);
		shrLog("Exiting...\n\n");
		exit(-1);
	}

	// Calculate dynamic size of memory buffers
	int vol_size = vol->npix * sizeof(float);
	int img_dim[2] = {cbi->dim[0], cbi->dim[1]};

	// Free cbi image
	free_cb_image(cbi);

	// Create global memory buffers on device
	cl_mem g_dev_vol[MAX_GPU_COUNT];

	// Create texture/image memory buffers on device
	size_t img_origin[3] = {0, 0, 0};
	size_t img_region[3] = {img_dim[0], img_dim[1], 1};
	size_t img_row_pitch = img_dim[0] * sizeof(float);

	cl_image_format img_format;
	img_format.image_channel_order = CL_R;
	img_format.image_channel_data_type = CL_FLOAT;

	cl_mem t_dev_img[MAX_GPU_COUNT];

	// constant memory buffers on device
	cl_mem c_dev_matrix[MAX_GPU_COUNT];
	cl_mem c_nrm[MAX_GPU_COUNT];
	cl_mem c_vol_offset[MAX_GPU_COUNT];
	cl_mem c_vol_pix_spacing[MAX_GPU_COUNT];
	cl_mem c_vol_dim[MAX_GPU_COUNT];
	cl_mem c_ic[MAX_GPU_COUNT];
	cl_mem c_img_dim[MAX_GPU_COUNT];
	cl_mem c_sad[MAX_GPU_COUNT];
	cl_mem c_scale[MAX_GPU_COUNT];
	cl_mem c_voxel_offset[MAX_GPU_COUNT];
	cl_mem c_voxel_device[MAX_GPU_COUNT];

	//kernel
	cl_kernel fdk_kernel[MAX_GPU_COUNT];

	for (cl_uint i = 0; i < device_count; i++) {
		// creates volume buffer
		g_dev_vol[i] = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, voxels_per_device[i].w * sizeof(float), NULL, &error);
		oclCheckError(error, CL_SUCCESS);

		// Create texture/image memory buffers on device
		t_dev_img[i] = clCreateImage2D(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, &img_format, img_dim[0], img_dim[1], 0, NULL, &error);
		oclCheckError(error, CL_SUCCESS);

		// Create constant memory buffers on device
		c_dev_matrix[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, matrix_size, NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		c_nrm[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, sizeof(float4), NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		c_vol_offset[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, sizeof(float4), NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		c_vol_pix_spacing[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, sizeof(float4), NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		c_vol_dim[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, sizeof(int4), NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		c_ic[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, sizeof(float2), NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		c_img_dim[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, sizeof(int2), NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		c_sad[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, sizeof(float), NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		c_scale[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, sizeof(float), NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		c_voxel_offset[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, sizeof(int4), NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		c_voxel_device[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, sizeof(int4), NULL, &error);
		oclCheckError(error, CL_SUCCESS);

		// Create the fdk kernel
		fdk_kernel[i] = clCreateKernel(program, "kernel_fdk", &error);
		oclCheckError(error, CL_SUCCESS);
	}
	
	// Wait for all queues to finish
	for (cl_uint i = 0; i < device_count; i++)
		clFinish(command_queue[i]);

	// Create variables for kernel and memory timing
	cl_event fdk_event[MAX_GPU_COUNT], hu_event[MAX_GPU_COUNT], img_event[MAX_GPU_COUNT], vol_event[MAX_GPU_COUNT];
	cl_ulong fdk_total = 0, hu_total = 0, img_total = 0, vol_total = 0;

	// Start working
	shrLog("Processing reconstruction of volume...\n\n");

	// Project each image into the volume one at a time
	for (image_num = options->first_img;  image_num <= options->last_img;  image_num += options->skip_img)	 {
		// Load the current image
		cbi = get_image(options, image_num);

		// Load dynamic kernel arguments
		kargs->img_dim.x = cbi->dim[0];
		kargs->img_dim.y = cbi->dim[1];
		kargs->ic.x = cbi->ic[0];
		kargs->ic.y = cbi->ic[1];
		kargs->nrm.x = cbi->nrm[0];
		kargs->nrm.y = cbi->nrm[1];
		kargs->nrm.z = cbi->nrm[2];
		kargs->sad = cbi->sad;
		kargs->sid = cbi->sid;

		for (int i = 0; i < 12; i++)
			kargs->matrix[i] = (float)cbi->matrix[i];

		// loop to copy data from host to each device
		for (cl_uint i = 0; i < device_count; i++) {
			kargs->v_offset = offset[i];
			kargs->v_device = voxels_per_device[i];

			// Copy texture/image memory from host to device
			error = clEnqueueWriteImage(command_queue[i], t_dev_img[i], CL_FALSE, img_origin, img_region, img_row_pitch, 0, cbi->img, 0, NULL, &img_event[i]);
			oclCheckError(error, CL_SUCCESS);

			// Copy constant memory from host to device
			error = clEnqueueWriteBuffer(command_queue[i], c_dev_matrix[i], CL_FALSE, 0, matrix_size, &kargs->matrix, 0, NULL, NULL);
			error |= clEnqueueWriteBuffer(command_queue[i], c_nrm[i], CL_FALSE, 0, sizeof(float4), &kargs->nrm, 0, NULL, NULL);
			error |= clEnqueueWriteBuffer(command_queue[i], c_vol_offset[i], CL_FALSE, 0, sizeof(float4), &kargs->vol_offset, 0, NULL, NULL);
			error |= clEnqueueWriteBuffer(command_queue[i], c_vol_pix_spacing[i], CL_FALSE, 0, sizeof(float4), &kargs->vol_pix_spacing, 0, NULL, NULL);
			error |= clEnqueueWriteBuffer(command_queue[i], c_vol_dim[i], CL_FALSE, 0, sizeof(int4), &kargs->vol_dim, 0, NULL, NULL);
			error |= clEnqueueWriteBuffer(command_queue[i], c_ic[i], CL_FALSE, 0, sizeof(float2), &kargs->ic, 0, NULL, NULL);		
			error |= clEnqueueWriteBuffer(command_queue[i], c_img_dim[i], CL_FALSE, 0, sizeof(int2), &kargs->img_dim, 0, NULL, NULL);
			error |= clEnqueueWriteBuffer(command_queue[i], c_sad[i], CL_FALSE, 0, sizeof(float), &kargs->sad, 0, NULL, NULL);
			error |= clEnqueueWriteBuffer(command_queue[i], c_scale[i], CL_FALSE, 0, sizeof(float), &kargs->scale, 0, NULL, NULL);
			error |= clEnqueueWriteBuffer(command_queue[i], c_voxel_offset[i], CL_FALSE, 0, sizeof(int4), &kargs->v_offset, 0, NULL, NULL);
			error |= clEnqueueWriteBuffer(command_queue[i], c_voxel_device[i], CL_FALSE, 0, sizeof(int4), &kargs->v_device, 0, NULL, NULL);
			oclCheckError(error, CL_SUCCESS);
		}

		// Wait for all queues to finish
		for (cl_uint i = 0; i < device_count; i++)
			clFinish(command_queue[i]);

		for (cl_uint i = 0; i < device_count; i++)
			img_total += executionTime(img_event[i]);

		for (cl_uint i = 0; i < device_count; i++) {
			// Invoke kernel
			// Sets Kernel Arguments
			error |= clSetKernelArg(fdk_kernel[i], 0, sizeof(cl_mem), (void *) &g_dev_vol[i]);
			error |= clSetKernelArg(fdk_kernel[i], 1, sizeof(cl_mem), (void *) &t_dev_img[i]);
			error |= clSetKernelArg(fdk_kernel[i], 2, sizeof(cl_mem), (void *) &c_dev_matrix[i]);
			error |= clSetKernelArg(fdk_kernel[i], 3, sizeof(cl_mem), (void *) &c_nrm[i]);
			error |= clSetKernelArg(fdk_kernel[i], 4, sizeof(cl_mem), (void *) &c_vol_offset[i]);
			error |= clSetKernelArg(fdk_kernel[i], 5, sizeof(cl_mem), (void *) &c_vol_pix_spacing[i]);
			error |= clSetKernelArg(fdk_kernel[i], 6, sizeof(cl_mem), (void *) &c_vol_dim[i]);
			error |= clSetKernelArg(fdk_kernel[i], 7, sizeof(cl_mem), (void *) &c_ic[i]);
			error |= clSetKernelArg(fdk_kernel[i], 8, sizeof(cl_mem), (void *) &c_img_dim[i]);
			error |= clSetKernelArg(fdk_kernel[i], 9, sizeof(cl_mem), (void *) &c_sad[i]);
			error |= clSetKernelArg(fdk_kernel[i], 10, sizeof(cl_mem), (void *) &c_scale[i]);
			error |= clSetKernelArg(fdk_kernel[i], 11, sizeof(cl_mem), (void *) &c_voxel_offset[i]);
			error |= clSetKernelArg(fdk_kernel[i], 12, sizeof(cl_mem), (void *) &c_voxel_device[i]);
			oclCheckError(error, CL_SUCCESS);
		}

		// Wait for all queues to finish
		for (cl_uint i = 0; i < device_count; i++)
			clFinish(command_queue[i]);

		// Invoke all kernels
		for (cl_uint i = 0; i < device_count; i++) {
			error = clEnqueueNDRangeKernel(command_queue[i], fdk_kernel[i], 3, NULL, fdk_global_work_size[i], fdk_local_work_size[i], 0, NULL, &fdk_event[i]);
			oclCheckError(error, CL_SUCCESS);
		}

		// Wait for fdk kernel to finish
		clWaitForEvents(device_count, fdk_event);

		for (cl_uint i = 0; i < device_count; i++)
			fdk_total += executionTime(fdk_event[i]);

		// Free the current image 
		free_cb_image(cbi);
	}

   	shrLog("Done reconstruction ...\n\n");
	
	// Start working
	shrLog("Processing conversion to Housfield scale...\n\n");

	size_t hu_local_work_size[MAX_GPU_COUNT];
	size_t hu_global_work_size[MAX_GPU_COUNT];

	cl_kernel hu_kernel[MAX_GPU_COUNT];
	for (cl_uint i = 0; i < device_count; i++) {
		// Creates the kernel object
		hu_kernel[i] = clCreateKernel(program, "convert_to_hu_cl", &error);
		oclCheckError(error, CL_SUCCESS);

		// Sets Kernel Arguments
		error = clSetKernelArg(hu_kernel[i], 0, sizeof(cl_mem), (void *) &g_dev_vol[i]);
		error |= clSetKernelArg(hu_kernel[i], 1, sizeof(int), &vol->npix);
		error |= clSetKernelArg(hu_kernel[i], 2, sizeof(int), &offset[i].w);
		error |= clSetKernelArg(hu_kernel[i], 3, sizeof(int), &voxels_per_device[i].w);
		oclCheckError(error, CL_SUCCESS);
	}

	// Wait for all queues to finish
	for (cl_uint i = 0; i < device_count; i++)
		clFinish(command_queue[i]);

	for (cl_uint i = 0; i < device_count; i++) {
		// Enqueue kernel and execute
		int hu_lBlock = 512;
		int hu_gBlock = voxels_per_device[i].w + (hu_lBlock - (voxels_per_device[i].w & (hu_lBlock - 1)));

		hu_local_work_size[i] = hu_lBlock;
		hu_global_work_size[i] = hu_gBlock;

		error = clEnqueueNDRangeKernel(command_queue[i], hu_kernel[i], 1, NULL, &hu_global_work_size[i], &hu_local_work_size[i], 0, NULL, &hu_event[i]);
		oclCheckError(error, CL_SUCCESS);
	}

	// Waits for hu kernel to finish
	clWaitForEvents(device_count, hu_event);

	shrLog("Done conversion ...\n\n");

	// Count kernel time
	for (cl_uint i = 0; i < device_count; i++)
		hu_total += executionTime(hu_event[i]);

	for (cl_uint i = 0; i < device_count; i++) {
		// Copy reconstructed volume from device to host
		error = clEnqueueReadBuffer(command_queue[i], g_dev_vol[i], CL_FALSE, 0, vol_size, (float*)vol->img + offset[i].w, 0, NULL, &vol_event[i]);
   		oclCheckError(error, CL_SUCCESS);
	}

	// Waits for volume to finish copying
	clWaitForEvents(device_count, vol_event);

	// Count host time
	for (cl_uint i = 0; i < device_count; i++)
		vol_total += executionTime(vol_event[i]);

	for (cl_uint i = 0; i < device_count; i++) {
		// Release kernels
		clReleaseKernel(fdk_kernel[i]);
		clReleaseKernel(hu_kernel[i]);

		// Release constant memory buffers
		clReleaseMemObject(c_voxel_device[i]);
		clReleaseMemObject(c_voxel_offset[i]);
		clReleaseMemObject(c_scale[i]);
		clReleaseMemObject(c_sad[i]);
		clReleaseMemObject(c_img_dim[i]);
		clReleaseMemObject(c_ic[i]);
		clReleaseMemObject(c_vol_dim[i]);
		clReleaseMemObject(c_vol_pix_spacing[i]);
		clReleaseMemObject(c_vol_offset[i]);
		clReleaseMemObject(c_nrm[i]);
		clReleaseMemObject(c_dev_matrix[i]);

		// Release texture/image memory buffers 
		clReleaseMemObject(t_dev_img[i]);
		
		// Release global memory buffers
		clReleaseMemObject(g_dev_vol[i]);	
	}

	// Perform timing
	float fdk_kernel_runtime = fdk_total * 1.0e-6f;
	float hu_kernel_runtime = hu_total * 1.0e-6f;
	float img_copy_runtime = img_total * 1.0e-6f;
	float vol_copy_runtime = vol_total * 1.0e-6f;
	float total_runtime = (fdk_kernel_runtime + hu_kernel_runtime + img_copy_runtime + vol_copy_runtime) * 1.0e-3f;

	shrLog("FDK Kernel run time: %f ms\n", fdk_kernel_runtime);
	shrLog("Hu Kernel run time: %f ms\n", hu_kernel_runtime);
	shrLog("CBI host to device copy run time: %f ms\n", img_copy_runtime);
	shrLog("Volume device to host copy run time: %f ms\n\n", vol_copy_runtime);

	return total_runtime;
}
